package main;

import game.Game;
import game.GameDynamics;
import game.Strategy;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Random;

import stats.UtilityObserver;
import util.Array;
import lemonade.ScaledLemonadeUtility;


/**
 * A main class for running the Lemonade game with three players.
 * @author maz
 *
 */
public class RunLemonadeGame {
  


  /**
   * Run one repeated lemonade game.
   * @param g the symmetric game to be played
   * @param strategyClassNames an array of class names for the strategies
   * @param options the options specified for the corresponding strategy classes
   * @param seedGenerator a seed generator for the classes
   * @return the utilities vector of the players
   * @throws IllegalArgumentException if the arguments to the constructor are not right (should not be thrown)
   * @throws InstantiationException if a class in strategyClassNames cannot be created
   * @throws IllegalAccessException if there is insufficient permission to construct a strategy
   * @throws InvocationTargetException if a class throws an exception during construction
   * @throws ClassNotFoundException if a class in strategyClassNames cannot be found
   * @throws NoSuchMethodException if the constructor for the class cannot be found
   */
  public static int[] oneRun(Game g, String[] strategyClassNames, String[] options, Random seedGenerator,int iterations) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    Strategy[] strategies = new Strategy[3];
    String paramString = "(game.Game,long,java.lang.String)";
    Class<?>[] params = new Class<?>[]{
        game.Game.class,
        long.class,String.class};
    int i=0;
    try{
      for(i=0;i<3;i++){
        Class<?> c = Class.forName(strategyClassNames[i]);
        Constructor<?> constr = c.getConstructor(params);
        strategies[i] = (Strategy)(constr.newInstance(g,seedGenerator.nextLong(),options[i]));
      }
    } catch(ClassNotFoundException cnf){
      System.err.println("Could not find "+strategyClassNames[i]);
      throw cnf;
    } catch(SecurityException s){
      System.err.println("No permission to access "+strategyClassNames[i]+paramString);
      throw s;
    } catch (NoSuchMethodException nsm){
      System.err.println("Could not find "+strategyClassNames[i]+paramString);
      throw nsm;
    } catch (InstantiationException ie){
      System.err.println("Cannot instantiate "+strategyClassNames[i]);
      throw ie;
    } catch (InvocationTargetException ite){
      System.err.println("Constructor "+strategyClassNames[i]+paramString+" threw an exception");
      throw ite;
    }
    
    GameDynamics dynamics = 
      new GameDynamics(g,strategies);
    UtilityObserver utilityObserver = new UtilityObserver(g.getUtility(),3);
    dynamics.addObserver(utilityObserver);
    dynamics.iterate(iterations);
    return utilityObserver.getUtilities();
  }
  
  public static int[] manyRuns(Game g, String[] strategies, String[] options, Random seedGenerator, int iterations, int resets) throws IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException{
    int[] totals = new int[3];
    for(int i=0;i<resets;i++){
      int[] utilities = oneRun(g,strategies, options,seedGenerator,iterations);
      for(int j=0;j<totals.length;j++){
        totals[j]+=utilities[j];
      }
    }
    return totals;
  }
  
  public static void showUsage(){
    System.err.println("Usage:RunLemonadeGame [-seed <SEED>|-numActions <ACTIONS>|-numRounds<ROUNDS> |-numRepeats <REPEATS>]* <P1CLASS> <P1OPTION> <P2CLASS> <P2OPTION> <P3CLASS> <P3OPTION>");
    System.exit(0);
  }
  
  public static void main(String[] args) throws ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException{
    
    if (args.length<6){
      showUsage();
    }
    long seed=0;
    int numActions=12;
    int iterations = 100;
    int repeats = 1000;
    
    int length=args.length-6;
    for(int i=0;i<length;i+=2){
      String arg = args[i];
      String fieldStr = args[i+1];
      long field = Long.parseLong(fieldStr);
      if (arg.equals("-seed")){
        seed = field;
      } else if (arg.equals("-actions")){
        if (100L<field){
          System.err.println("Number of actions needs to be less than 100");
        }
        numActions = (int)field;
      } else if (arg.equals("-rounds")){
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of actions needs to be less than 100");
        }
        iterations = (int)field;
      } else if (arg.equals("-repeats")){
        if ((long)(Integer.MAX_VALUE)<field){
          System.err.println("Number of actions needs to be less than 100");
        }
        repeats = (int)field;
      }
    }
    
    Random seedGenerator = new Random(seed);
    ScaledLemonadeUtility utility = new ScaledLemonadeUtility(numActions);
    Game g = new Game(new int[]{numActions,numActions,numActions},utility);

    String[] strategies = new String[3];
    String[] options = new String[3];
    
    int playerIndex = 0;
    for(int i=args.length-6;i<args.length;i+=2){
      strategies[playerIndex]=args[i];
      options[playerIndex]=args[i+1];
      playerIndex++;
    }
    
    System.out.println(Array.arrayToNakedString(strategies));
    int[] totals = manyRuns(g,strategies,options,seedGenerator,iterations,repeats);
    System.out.println(Array.arrayToString(totals));
    
  }
}
